"""
Defining Job class
"""
from typing_extensions import Union
import requests

import json
import lxml.etree as ET

from .utils import id_from_info, get_text_key_vals, check_wildcard, parse_wildcard


PARAMETER_TAGS = ['integer','float','double','boolean','string','integer-vector','float-vector','double-vector','string-vector',
                'integer-enumeration','float-enumeration','double-enumeration','string-enumeration','file','directory','image',
                'geometry','point','pointfile','region','table','transform']

JOB_STATUS_KEY = [
    'INACTIVE',
    'QUEUED',
    'RUNNING',
    'SUCCESS',
    'ERROR',
    'CANCELED'
]


class Job:
    """Base class of Job
    """
    def __init__(self,
                 gc,
                 plugin_id:Union[str,None] = None,
                 docker_image: Union[str,None] = None,
                 cli: Union[str,None] = None,
                 input_args: Union[list,None] = None
                 ):
        
        self.gc = gc
        self.plugin_id = plugin_id
        self.docker_image = docker_image
        self.cli = cli
        self.input_args = input_args
        self.job_id = None

        # Either id is defined or both docker_image and cli have to be defined
        assert any([not self.plugin_id is None, all([not j is None for j in [self.docker_image, self.cli]])])

        self.executable_dict = self.get_plugin_info()
        self.inputs = self.parse_input_args()


    def get_plugin_info(self):
        """Finding the plugin in the provided DSA instance either by id or combination of docker_image and cli_name
        """
        executable_dict = None
        if self.plugin_id is None:
            # Already asserted, if self.job_id is None then docker_image and cli have to be defined
            plugin_info = id_from_info(self.gc, self.docker_image, self.cli)

            if not plugin_info is None:
                self.plugin_id = plugin_info['_id']
                executable_dict = self.get_executable()
        else:
            executable_dict = self.get_executable()


        return executable_dict

    def get_executable(self)->dict:
        """Create the executable dictionary for this plugin. Finds plugin descriptive information and organizes inputs in a ready-to-use format.

        :return: Executable dictionary with descriptive information, parameter groups, and each parameter groups inputs
        :rtype: dict
        """
        executable_dict = None
        if not self.plugin_id is None:
            plugin_xml_req = requests.get(
                self.gc.urlBase+f'slicer_cli_web/cli/{self.plugin_id}/xml?token={self.gc.token}'
            )

            if plugin_xml_req.status_code==200:
                plugin_xml = ET.fromstring(plugin_xml_req.content)

                executable_dict = {
                    'title': plugin_xml.find('title'),
                    'description': plugin_xml.find('description'),
                    'author': plugin_xml.find('contributor'),
                    'documentation': plugin_xml.find('documentation-url')
                }

                executable_dict = get_text_key_vals(executable_dict)

                parameters_list = []
                for param in plugin_xml.iterfind('parameters'):
                    param_dict = {
                        'advanced': param.find('advanced'),
                        'label': param.find('label'),
                        'description': param.find('description')
                    }

                    param_dict = get_text_key_vals(param_dict)

                    input_list = []
                    for sub_el in param:
                        if sub_el.tag in PARAMETER_TAGS:
                            input_dict = {
                                'type': sub_el.tag,
                                'label': sub_el.find('label'),
                                'name': sub_el.find('name'),
                                'channel': sub_el.find('channel'),
                                'description': sub_el.find('description'),
                                'default': sub_el.find('default')
                            }
                            input_dict = get_text_key_vals(input_dict)

                            if 'enumeration' in sub_el.tag:
                                options_list = []
                                for opt in sub_el.iterfind('element'):
                                    options_list.append(opt.text)

                                input_dict['options'] = options_list
                        
                            if not sub_el.find('constraints') is None:
                                constraints = sub_el.find('constraints')
                                if not constraints is None:
                                    constraints_dict = {
                                        'min': constraints.find('min'),
                                        'max': constraints.find('max'),
                                        'step': constraints.find('step')
                                    }
                                    constraints_dict = get_text_key_vals(constraints_dict)

                                    input_dict['constraints'] = constraints_dict

                            input_list.append(input_dict)

                    param_dict['inputs'] = input_list

                    parameters_list.append(param_dict)

                executable_dict['parameters'] = parameters_list

            else:
                print(plugin_xml_req.content)

        return executable_dict

    def get_defaults(self)->list:
        """Method for finding all of the default values for the current plugin

        :return: List of {'name': 'default'} or {'label': 'default'} values for each input
        :rtype: list
        """
        defaults_list = []
        if not self.executable_dict is None:
            for p in self.executable_dict['parameters']:
                for i in p['inputs']:
                    defaults_list.append({
                        'name': i['name'],
                        'default': i['default']
                    })

        return defaults_list

    def find_input(self, input_name):

        exe_input = None
        for p in self.executable_dict['parameters']:
            for inp in p['inputs']:
                if inp['name']==input_name:
                    exe_input = inp
                    break
        
        return exe_input

    def parse_input_args(self):
        """Method for organizing user-provided job input values. Only non-default valued inputs are required.
        """

        if not self.input_args is None:
            user_input_names = [i['name'] for i in self.input_args]
            
            for i in self.input_args:
                if type(i['value'])==str:
                    if check_wildcard(i['value']):
                        i['value'] = parse_wildcard(self.gc, i['value'])
        
        else:
            user_input_names = []

        inputs_list = []
        input_names = []
        # Replacing default values
        plugin_defaults = self.get_defaults()
        print(plugin_defaults)
        for p_d in plugin_defaults:
            if 'name' in p_d:
                if p_d['name'] in user_input_names:
                    inputs_list.append({
                        'name': p_d['name'],
                        'value': self.input_args[user_input_names.index(p_d['name'])]['value']
                    })
                else:
                    inputs_list.append({
                        'name': p_d['name'],
                        'value': p_d['default']
                    })
                
                input_names.append(p_d['name'])
            elif 'label' in p_d:
                # This is for non-named inputs which are defined by index
                # Ideally, every input parameter would have a name for reference but some may just have
                # an index.
                inputs_list.append({
                    'name': None,
                    'value': p_d['default']
                })
                input_names.append(p_d['label'])

        # Non-default value-having inputs have to be defined in input_args or they'll be missing
        if not self.input_args is None:
            for i in self.input_args:
                if not i['name'] in input_names:
                    inputs_list.append({
                        'name': i['name'],
                        'value': i['value']
                    })
        
        # Adding girderApiUrl and girderToken
        inputs_list.extend([
            {
                'name': 'girderApiUrl',
                'value': self.gc.urlBase
            },
            {
                'name': 'girderToken',
                'value': self.gc.token
            }
        ])
        input_names.extend(['girderApiUrl','girderToken'])

        # Checking that all inputs are present
        required_inputs = []
        for p in self.executable_dict['parameters']:
            for i in p['inputs']:
                if not i['name'] is None:
                    required_inputs.append(i['name'])
                elif not i['label'] is None:
                    required_inputs.append(i['label'])

        if not all([i in input_names for i in required_inputs]):
            print(f'Missing required inputs')
            for i in required_inputs:
                if not i in input_names:
                    print(f'{i} is missing')
            
            print(f'Provided inputs: {input_names}')


        return inputs_list

    def cancel(self):
        """Send cancel request for this job
        """

        if not self.job_id is None:
            cancel_response = self.gc.put(
                f'/job/{self.job_id}/cancel'
            )

            return cancel_response
        else:
            return {'message': 'Job has not started yet'}

    def start(self):
        """Send start request for this job
        """
        
        start_request = requests.post(
            url = self.gc.urlBase+f'slicer_cli_web/cli/{self.plugin_id}/run?token={self.gc.token}',
            params = {
                i['name']: i['value']
                for i in self.inputs
            }
        )

        if start_request.status_code==200:
            self.job_info = start_request.json()
            self.job_id = self.job_info['_id']

        return start_request

    def get_status(self):
        """Get the status of the current job
        """

        if not self.job_id is None:
            job_info = self.gc.get(f'/job/{self.job_id}')
            job_status_idx = job_info['status']

            return JOB_STATUS_KEY[job_status_idx]
        else:
            return JOB_STATUS_KEY[0]

    def get_logs(self):
        """Get logs of this job
        """

        if not self.job_id is None:
            job_info = self.gc.get(f'/job/{self.job_id}')
            job_logs = job_info['log']

            log_list = [i.split('\n') for i in job_logs]
        else:
            lob_list = ['Job has not started yet']

        return log_list






"""Defining Sequence class
"""

from typing_extensions import Union
import requests
from time import sleep

from .utils import get_unique_id

class Sequence:
    """Base class of Sequence, containing multiple jobs
    """
    def __init__(self,
                 gc,
                 jobs: list = []):
        
        self.gc = gc
        self.jobs = jobs
        self.id = get_unique_id()

    def get_logs(self, type = 'all'):
        
        assert type in ['all','running','finished']
        
        current_job_statuses = self.get_status()

        logs_list = []
        if type=='all':
            for j,status in zip(self.jobs,current_job_statuses):
                if status not in ['INACTIVE','QUEUED']:
                    logs_list.append(
                        {
                            'Job Name': j.executable_dict.title,
                            'Logs': j.get_logs()
                        }
                    )
        elif type=='running':
            for j,status in zip(self.jobs,current_job_statuses):
                if status=='RUNNING':
                    logs_list.append(
                        {
                            'Job Name': j.executable_dict.title,
                            'Logs': j.get_logs()
                        }
                    )
        
        elif type == 'finished':
            for j,status in zip(self.jobs,current_job_statuses):
                if status in ['SUCCESS','ERROR','CANCELED']:
                    logs_list.append(
                        {
                            'Job Name': j.executable_dict.title,
                            'Logs': j.get_logs()
                        }
                    )

        return logs_list

    def get_status(self)->list:
        """Get the statuses of all jobs in a sequence

        :return: List of {'Job Name': '', 'Status': ''} dictionaries
        :rtype: list
        """
        status_list = []
        for j in self.jobs:
            status_list.append(
                {'Job Name': j.executable_dict["title"], 'Status': j.get_status()}
            )

        return status_list

    def cancel(self, type:str = 'all')->list:
        """Cancel either all, running, queued, or inactive jobs in a sequence

        :param type: str, defaults to 'all'
        :type type: str, optional
        :return: Either a list (if multiple jobs are canceled) or a single dictionary with the cancellation response
        :rtype: list
        """

        assert type in ['all','running','queued','inactive']
        
        cancel_responses = []
        current_job_statuses = self.get_status()
        if type =='all':
            for j,status in zip(self.jobs,current_job_statuses):
                if not status in ['SUCCESS','ERROR','CANCELED']:
                    cancel_responses.append(j.cancel())
        
        else:
            for j,status in zip(self.jobs,current_job_statuses):
                if status==type.upper():
                    cancel_responses.append(j.cancel())

        return cancel_responses

    def add_sequence_metadata(self, job, job_idx):

        # This might not actually be possible to add
        put_response = self.gc.put(
            f'/job/{job.job_id}/metadata',
            data = {
                'part_of': self.id,
                'preceded_by': self.jobs[job_idx-1].job_id if job_idx>0 else '',
                'followed_by': self.jobs[job_idx+1].plugin_id if len(self.jobs)>job_idx+1 else ''
            }
        )

        return put_response

    def start(self, check_interval:int = 5, cancel_on_error:bool = True,verbose:bool = False):
        """Start the job sequence, checking the status of running jobs every "check_interval" seconds

        :param check_interval: How many seconds to go between status checks, defaults to 5
        :type check_interval: int, optional
        :param cancel_on_error: Whether to cancel the whole job sequence if one of the jobs fails, defaults to True
        :type cancel_on_error: bool, optional
        :param verbose: Whether to print current job and status at each check
        :type verbose: bool, optional
        """

        assert check_interval>0

        for job_idx, job in enumerate(self.jobs):

            job_request = job.start()
            if job_request.status_code==200:
                #job_info = job_request.json()
                current_status = job.get_status()
                #self.add_sequence_metadata(job,job_idx)

                while not current_status in ['SUCCESS','ERROR','CANCELED']:
                    sleep(check_interval)

                    current_status = job.get_status()

                    if verbose:
                        print('-------------------------')
                        print(f'On {job.executable_dict["title"]}, Status: {current_status}')
                        print('-------------------------')


                    if current_status in ['ERROR','CANCELED']:
                        
                        if verbose:
                            print('XXXXXXXXXXXXXXXXXXXXXXXXXXXX')
                            print(f'{current_status} encountered on job: {job.job_id}, {job.executable_dict["title"]}')
                            print('XXXXXXXXXXXXXXXXXXXXXXXXXXXX')

                        if cancel_on_error:
                            if verbose:
                                print('Canceling remaining jobs in sequence')

                            self.cancel()
            else:

                print('Error submitting job request')
                print(f'Status Code: {job_request.status_code}')
                print(job_request.content)

                if cancel_on_error:
                    self.cancel()










"""Utility functions used by girder-job-sequence
"""

import os
from typing_extensions import Union
import json

from uuid import uuid4

def get_unique_id():
    """Create a unique id for something"""
    return uuid4().hex[:24]

def id_from_info(gc, docker_image_name: str, cli_name: str):
    """Find a plugin's info from the name of the Docker image (image/name:tag) and CLI name

    :param gc: Girder client handler
    :type gc: None
    :param docker_image_name: Name of the Docker image that has that CLI (image/name:tag)
    :type docker_image_name: str
    :param cli_name: Name of the CLI
    :type cli_name: str
    """

    plugin_list = gc.get('/slicer_cli_web/cli')

    plugin_info = None
    for p in plugin_list:
        if p['image']==docker_image_name and p['name']==cli_name:
            plugin_info = p
            break

    return plugin_info

def get_text_key_vals(xml_dict:dict):
    """Used for grabbing ".text" of values in dictionary containing XML sub-elements

    :param xml_dict: Dictionary containing keys and values with the "text" method or None
    :type xml_dict: dict
    :return: Same dictionary but with XML values replaced with their "text"
    :rtype: dict
    """
    return_dict = xml_dict.copy()
    for key,val in return_dict.items():
        if not val is None:
            try:
                return_dict[key] = val.text
            except:
                return_dict[key] = val
    
    return return_dict

def find_item(gc,type: str, query:str):

    if type=='path':
        item_info = gc.get(f'/resource/lookup',parameters={'path': query})['_id']
    elif type=='_id':
        item_info = gc.get(f'/item/{query}')['_id']

    return item_info

def find_file(gc, item_type:str, item_query:str, file_type:str, file_query:str):

    if item_type == 'path':
        item_info = find_item(gc,item_type, item_query)
    elif item_type=='_id':
        item_info = item_query
    
    if file_type == 'fileName':
        item_files = gc.get(f'/item/{item_info}/files',parameters = {'limit': 0})

        file_names = [i['name'] for i in item_files]
        file_info = item_files[file_names.index(file_query)]['_id']
    
    elif file_type == '_id':
        file_info = file_query

    return file_info

def find_annotation(gc, item_type, item_query, annotation_type, annotation_query):

    if item_type=='path': 
        item_info = find_item(gc, item_type,item_query)
    elif item_type == '_id':
        item_info = item_query

    if annotation_type == 'annotationName':
        item_annotations = gc.get(f'/annotation',parameters={'itemId': item_info["_id"]})

        annotation_names = [i['annotation']['name'] for i in item_annotations]
        annotation_info = item_annotations[annotation_names.index(annotation_query)]['_id']
    elif annotation_type=='annotationId':
        annotation_info = annotation_query

    return annotation_info

def check_wildcard(test_str: str):
    """Check whether a string is a candidate for wildcard input

    :param test_str: String value to test
    :type test_str: str
    """
    return '{{' in test_str

def parse_wildcard(gc, wildcard_str:str):
    """Parse wildcard input, using type, {item,file,or annotation}_type and {item,file,or annotation}_query key-val pairs to search for items, files, or annotations

    :param gc: Girder client handler
    :type gc: None
    :param wildcard_str: String containing "{{}}" wildcard indicator
    :type wildcard_str: str
    """
    # Verifying this is a wildcard candidate
    assert '{{' in wildcard_str
    wildcard_args = json.loads(wildcard_str[1:-1].replace("'",'"'))

    if wildcard_args['type']=='item':
        wildcard_val = find_item(gc, wildcard_args['item_type'], wildcard_args['item_query'])
    elif wildcard_args['type']=='folder':
        # This is the same function since they both use the /resource/lookup endpoint for path types
        wildcard_val = find_item(gc,wildcard_args['folder_type'],wildcard_args['folder_query'])
    elif wildcard_args['type']=='file':
        wildcard_val = find_file(gc, wildcard_args['item_type'],wildcard_args['item_query'],wildcard_args['file_type'],wildcard_args['file_query'])
    elif wildcard_args['type']=='annotation':
        wildcard_val = find_annotation(gc,wildcard_args['item_type'],wildcard_args['item_query'],wildcard_args['annotation_type'],wildcard_args['anotation_query'])

    return wildcard_val

def from_json(gc, json_path: str):
    """Read job or job sequence from JSON file

    :param gc: Girder client handler
    :type gc: None
    :param json_path: Path to json file containing data
    :type json_path: str
    """

    assert os.path.exists(json_path)

    with open(json_path,'r') as f:
        job_json = json.load(f)

        f.close()
    
    if type(job_json)==dict:
        job_json = [job_json]
    
    job_list = from_list(job_json)
    
    return job_list

def from_dict(gc, dict_data:dict):
    """Method for creating job/sequence from a dictionary

    :param gc: Girder client handler
    :type gc: None
    :param dict_data: _description_
    :type dict_data: dict
    """
    from .job import Job

    job_from_dict = Job(
        gc = gc,
        plugin_id = dict_data['plugin_id'] if 'plugin_id' in dict_data else None,
        docker_image=dict_data['docker_image'] if 'docker_image' in dict_data else None,
        cli= dict_data['cli'] if 'cli' in dict_data else None,
        input_args = dict_data['input_args'] if 'input_args' in dict_data else None
    )

    return job_from_dict

def from_list(gc, list_data:list):
    """Generating job/sequence from list of dictionaries

    :param list_data: _description_
    :type list_data: list
    """

    job_list = []
    for l in list_data:
        job_list.append(
            from_dict(gc,l)
        )

    if len(job_list)>1:
        from .sequence import Sequence
        return Sequence(gc,job_list)
    else:
        return job_list[0]





